/*****************************************************************************
 * davinci.c: decoder and encoder modules using the DaVinci DSP.
 *****************************************************************************
 * Copyright (C) 2008 M2X BV
 * $Id: 791f28707d0a471ed1538148879e7360a4bef58b $
 *
 * Authors: Antoine Cellerier <dionoea at videolan dot org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston MA 02110-1301, USA.
 *****************************************************************************/

/*****************************************************************************
 * Preamble
 *****************************************************************************/
#include "davinci.h"

/*****************************************************************************
 * Module descriptor
 *****************************************************************************/
#define VIDDEC_CHROMA_TEXT N_( "Decoder output chroma" )
#define VIDDEC_CHROMA_LONGTEXT N_( \
    "Force video decoder to output in specific chroma" )

vlc_module_begin();
    set_category( CAT_INPUT );
    set_subcategory( SUBCAT_INPUT_VCODEC );
    set_description( _("DaVinci DSP video decoder") );
    set_capability( "decoder", 1337 );
    set_callbacks( OpenVideoDecoder, CloseVideoDecoder );
    add_string( "davinci-viddec-chroma", "", NULL,
                VIDDEC_CHROMA_TEXT, VIDDEC_CHROMA_LONGTEXT, VLC_TRUE );

    add_submodule();
    set_description( _("DaVinci DSP audio decoder") );
    set_capability( "decoder", 0 );
    set_callbacks( OpenAudioDecoder, CloseAudioDecoder );

    add_submodule();
    set_description( _("DaVinci DSP video encoder" ) );
    set_capability( "encoder", 1337 );
    set_callbacks( OpenVideoEncoder, CloseVideoEncoder );
vlc_module_end();


/*****************************************************************************
 * Memory utils
 *****************************************************************************/
#include <ti/sdo/ce/osal/Memory.h>

int __AllocateBuffer( vlc_object_t *p_this, XDAS_Int32 i_num,
                      XDAS_Int32 *pi_sizes, XDM_BufDesc *buf )
{
    int i;

    buf->numBufs = i_num;

    buf->bufs = (XDAS_Int8 **)malloc( buf->numBufs * sizeof( XDAS_Int8 * ) );
    if( !buf->bufs )
    {
        buf->numBufs = 0;
        return VLC_ENOMEM;
    }

    buf->bufSizes = (XDAS_Int32 *)malloc( buf->numBufs * sizeof( XDAS_Int32 ) );
    if( !buf->bufSizes )
    {
        free( buf->bufs );
        buf->numBufs = 0;
        return VLC_ENOMEM;
    }

    for( i = 0; i < i_num; i++ )
    {
        buf->bufSizes[i] = pi_sizes[i];
        buf->bufs[i] = Memory_contigAlloc( pi_sizes[i], Memory_DEFAULTALIGNMENT );
        if( !buf->bufs[i] )
        {
            msg_Err( p_this, "Failed to allocate buffer" );
            for( i--; i >= 0; i-- )
                Memory_contigFree( buf->bufs[i], buf->bufSizes[i] );

            free( buf->bufs );
            free( buf->bufSizes );
            buf->numBufs = 0;
            return VLC_ENOMEM;
        }
    }
    return VLC_SUCCESS;
}

void FreeBuffer( XDM_BufDesc *buf )
{
    if( buf->numBufs != 0 )
    {
        int i;
        for( i = 0; i < buf->numBufs; i++ )
            Memory_contigFree( buf->bufs[i], buf->bufSizes[i] );
        free( buf->bufs );
        free( buf->bufSizes );
        buf->numBufs = 0;
    }
}

/*****************************************************************************
 * Misc utils
 *****************************************************************************/
const char *ppsz_engine_error[] = {
    [Engine_EOK] = "Ok",
    [Engine_EEXIST] = "Engine name doesn't exist",
    [Engine_ENOMEM] = "Can't allocate engine memory",
    [Engine_EDSPLOAD] = "Unable to load the DSP image. This probably means that you are already runing a different engine on the DSP. Since the DSP doesn't support multiple engines at the same time this will fail. If you have another engine available with support for all the encoders and decoders that you will need simultaneously, force it's use with the --davinci-*-engine command line options. See vlc -p davinci --advanced --help-verbose for information about those command line options.",
    [Engine_ENOCOMM] = "Can't create communication connection to DSP",
    [Engine_ENOSERVER] = "Can't locate the server on the DSP",
    [Engine_ECOMALLOC] = "Can't allocate communication buffer",
    [Engine_ERUNTIME] = "Engine runtime failure",
    [Engine_ECODECCREATE] = "Engine codec creation failed",
    [Engine_ECODECSTART] = "Engine codec start failed",
    [Engine_EINVAL] = "Bad parameter",
    [Engine_EBADSERVER] = "Incompatible server specified",
    [Engine_ENOTAVAIL] = "Service not available",
    //[Engine_EWRONGSTATE] = "Call can't be made at this time",
    //[Engine_EINUSE] = "Call can't be made at this time because a required resource is in use",
    //[Engine_ENOTFOUND] = "Entity was not found",
};

const char * davinci_GetExtendedError( XDAS_Int32 error )
{
    if( XDM_ISAPPLIEDCONCEALMENT( error ) )
        return "Applied bit concealment";
    if( XDM_ISINSUFFICIENTDATA( error ) )
        return "Insufficient data";
    if( XDM_ISCORRUPTEDDATA( error ) )
        return "Corrupted data";
    if( XDM_ISCORRUPTEDHEADER( error ) )
        return "Corrupted header";
    if( XDM_ISUNSUPPORTEDINPUT( error ) )
        return "Unsupported input";
    if( XDM_ISUNSUPPORTEDPARAM( error ) )
        return "Unsupported param";
    if( XDM_ISFATALERROR( error ) )
        return "Fatal error";

    return "";
}

void __PrintAvailableAlgorithms( vlc_object_t *p_this, const char *psz_engine )
{
    int i = 0;
    Engine_AlgInfo info;
    info.algInfoSize = sizeof( info );
    msg_Dbg( p_this, "Available algorithms in engine `%s':", psz_engine );
    while( Engine_getAlgInfo( (String)psz_engine, &info, i ) == Engine_EOK )
    {
        const char **psz_type = (const char **)info.typeTab;
        i++;
        msg_Dbg( p_this, "    %d: %s (%s)", i, info.name, info.isLocal?"local":"remote" );
        if( *psz_type )
        {
            msg_Dbg( p_this, "        Inheritance hierarchy:" );
            for( ; *psz_type; psz_type++ )
                msg_Dbg( p_this, "            %s", *psz_type );
        }
    }
}

XDAS_Int32 VlcChromaToXdm( vlc_fourcc_t i_chroma )
{
    switch( i_chroma )
    {
        case VLC_FOURCC('I','4','2','0'):
            return XDM_YUV_420P;

        case VLC_FOURCC('I','4','1','1'):
            return XDM_YUV_411P;

        case VLC_FOURCC('I','4','2','2'):
            return XDM_YUV_422P;

        case VLC_FOURCC('I','4','4','4'):
            return XDM_YUV_444P;

        case VLC_FOURCC('Y','U','Y','V'):
            return XDM_YUV_422IBE; /* FIXME ? */

        case VLC_FOURCC('U','Y','V','Y'):
            return XDM_YUV_422ILE; /* FIXME ? */

        case VLC_FOURCC('G','R','E','Y'):
            return XDM_GRAY;

#if 0
        /* FIXME: not sure about this one */
        case VLC_FOURCC('R','V','2','4'):
        case VLC_FOURCC('R','V','3','2'):
            return XDM_RGB;
#endif

        default:
            return XDM_CHROMA_NA;
    }
}

