/*****************************************************************************
 * viddec.c: video decoder module using the DaVinci DSP.
 *****************************************************************************
 * Copyright (C) 2008 M2X BV
 * $Id: be352bd11a8eb10b9f1bb5815737666a140102d0 $
 *
 * Authors: Antoine Cellerier <dionoea at videolan dot org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston MA 02110-1301, USA.
 *****************************************************************************/

/*****************************************************************************
 * Preamble
 *****************************************************************************/
#include "davinci.h"
#include <assert.h>
#include <ti/sdo/ce/video/viddec.h>

/*****************************************************************************
 * Local prototypes
 *****************************************************************************/
static picture_t *DecodeVideoBlock( decoder_t *, block_t ** );
static picture_t *DecodeVideoBlockInner( decoder_t *, block_t **, int );

/*****************************************************************************
 * Video decoder
 *****************************************************************************/
struct decoder_sys_t
{
    Engine_Handle e;
    VIDDEC_Handle d;

    XDM_BufDesc in;
    XDM_BufDesc out;
};

/*****************************************************************************
 * OpenDecoder: probe the decoder and return score
 *****************************************************************************/
int OpenVideoDecoder( vlc_object_t *p_this )
{
    decoder_t *p_dec = (decoder_t*)p_this;
    decoder_sys_t *p_sys;
    Engine_Error err;
    char *psz_chroma = NULL;
    char *psz_codec = NULL;
    vlc_fourcc_t i_chroma = VLC_FOURCC('U','Y','V','Y');
    VIDDEC_Params params;

    switch( p_dec->fmt_in.i_codec )
    {
#if 0
        CASE_MPEG1
            psz_codec = strdup( "mpegdec" ); /* FIXME */
            break;
#endif

        CASE_MPEG2
            psz_codec = strdup( "mpeg2dec" );
            break;

        CASE_MPEG4
            psz_codec = strdup( "mpeg4dec" );
            break;

        CASE_H264
            psz_codec = strdup( "h264dec" );
            break;

#if 0
            CASE_VC1
                psz_codec = strdup( "vc1dec" ); /* FIXME */
                break;
#endif

        default:
            return VLC_EGENERIC;
    }

    /* Allocate our private structure */
    p_dec->p_sys = (decoder_sys_t *)malloc( sizeof( decoder_sys_t ) );
    if( !p_dec->p_sys )
        return VLC_ENOMEM;
    p_sys = p_dec->p_sys;
    memset( p_sys, 0, sizeof( decoder_sys_t ) );

    /* Initialize the codec engine */
    CERuntime_init();

    /* Create an engine handle */
    #define ti_engine "decode"
    p_sys->e = Engine_open( ti_engine, NULL /*&Engine_ATTRS*/, &err );
    if( err != Engine_EOK )
    {
        msg_Err( p_dec, "Error while opening engine `%s': %s",
                 ti_engine, ppsz_engine_error[err] );
        goto error;
    }

    PrintAvailableAlgorithms( p_this, ti_engine );

    /* Get user supplied chroma setting */
    psz_chroma = config_GetPsz( p_dec, "davinci-viddec-chroma" );
    if( psz_chroma && strlen( psz_chroma ) == 4 )
    {
        i_chroma = VLC_FOURCC( psz_chroma[0], psz_chroma[1],
                               psz_chroma[2], psz_chroma[3] );
        msg_Dbg( p_dec, "Forcing output chroma to `%s'", psz_chroma );
    }
    free( psz_chroma );

    /* Create video decoder */
    params.size = sizeof( VIDDEC_Params );

    /* FIXME: put maximum settings accepted by the hardware */
    params.maxHeight = p_dec->fmt_in.video.i_height ?: 720;
    params.maxWidth = p_dec->fmt_in.video.i_width ?: 1280;
    params.maxFrameRate = 25000; /* in frames per 1000 seconds */
    params.maxBitRate = 1000000; /* in bits per second */

    params.dataEndianness = XDM_BYTE;
    params.forceChromaFormat = VlcChromaToXdm( i_chroma );
    if( params.forceChromaFormat == XDM_CHROMA_NA )
    {
        msg_Err( p_dec, "Unsupported output chroma (%4.4s).",
                 (const char *)&i_chroma );
        goto error;
    }

    p_sys->d = VIDDEC_create( p_sys->e, (String)psz_codec, &params );
    if( !p_sys->d )
    {
        msg_Err( p_dec, "Failed to create video decoder (%s)", psz_codec );
        goto error;
    }

    /* Set output properties */
    p_dec->fmt_out.i_cat = VIDEO_ES;
    p_dec->fmt_out.i_codec =
    p_dec->fmt_out.video.i_chroma = i_chroma;
    p_dec->fmt_out.video.i_width = p_dec->fmt_in.video.i_width;
    p_dec->fmt_out.video.i_height = p_dec->fmt_in.video.i_height;
    p_dec->fmt_out.video.i_aspect = p_dec->fmt_in.video.i_aspect;

    /* Set callbacks */
    p_dec->pf_decode_video = DecodeVideoBlock;

    free( psz_codec );
    return VLC_SUCCESS;

error:
    if( p_sys->e ) Engine_close( p_sys->e );
    free( p_sys );
    free( psz_codec );
    return VLC_EGENERIC;
}

/*****************************************************************************
 * CloseDecoder: png decoder destruction
 *****************************************************************************/
void CloseVideoDecoder( vlc_object_t *p_this )
{
    decoder_t *p_dec = (decoder_t *)p_this;
    decoder_sys_t *p_sys = p_dec->p_sys;

    /* Close our codec handle */
    VIDDEC_delete( p_sys->d );

    /* Close our engine handle */
    Engine_close( p_sys->e );

    /* Exit the codec engine */
    CERuntime_exit();

    /* Free 'DaVinci compliant' buffers */
    FreeBuffer( &p_sys->in );
    FreeBuffer( &p_sys->out );

    free( p_sys );
}

/****************************************************************************
 * DecodeBlock: the whole thing
 ****************************************************************************/
static picture_t *DecodeVideoBlock( decoder_t *p_dec, block_t **pp_block )
{
    return DecodeVideoBlockInner( p_dec, pp_block, 0 );
}

static picture_t *DecodeVideoBlockInner( decoder_t *p_dec, block_t **pp_block, int i_extra )
{
    decoder_sys_t *p_sys = p_dec->p_sys;
    block_t *p_block;
    picture_t *p_pic = NULL;
    VIDDEC_InArgs in_args;
    VIDDEC_OutArgs out_args;
    VIDDEC_DynamicParams dparams;
    VIDDEC_Status status;
    int i;

    if( !pp_block || !*pp_block ) return NULL;
    p_block = *pp_block;

    if( p_block->i_flags & (BLOCK_FLAG_DISCONTINUITY|BLOCK_FLAG_CORRUPTED) )
    {
        block_Release( p_block );
        return NULL;
    }

    memset( &in_args, 0, sizeof( in_args ) );
    memset( &out_args, 0, sizeof( out_args ) );

    /* Configure video decoder */
    dparams.size = sizeof( VIDDEC_DynamicParams );
    memset( &status, 0, sizeof( VIDDEC_Status ) );
    status.size = sizeof( VIDDEC_Status );

    if( p_sys->in.numBufs == 0 || p_sys->out.numBufs == 0 )
    {
        if( VIDDEC_control( p_sys->d, XDM_GETBUFINFO, &dparams, &status )
            != VIDDEC_EOK )
        {
            msg_Err( p_dec, "Failed to get buffer info" );
            goto error;
        }

        /* Allocate input buffer */
        if( AllocateBuffer( p_dec, status.bufInfo.minNumInBufs,
                            status.bufInfo.minInBufSize, &p_sys->in )
            != VLC_SUCCESS )
        {
            msg_Err( p_dec, "Failed to allocate input buffers" );
            goto error;
        }

        /* Allocate output buffer(s) */
        if( AllocateBuffer( p_dec, status.bufInfo.minNumOutBufs,
                            status.bufInfo.minOutBufSize, &p_sys->out )
            != VLC_SUCCESS )
        {
            FreeBuffer( &p_sys->in );
            msg_Err( p_dec, "Failed to allocate output buffers" );
            goto error;
        }
    }

    if( VIDDEC_control( p_sys->d, XDM_GETSTATUS, &dparams, &status )
        != VIDDEC_EOK )
    {
        msg_Err( p_dec, "Failed to get decoder status" );
        goto error;
    }

    /* Setup input arguments */
    in_args.size = sizeof( in_args );
    in_args.numBytes = __MIN( i_extra + p_block->i_buffer,
                              p_sys->in.bufSizes[0] );
    in_args.inputID = 1;

    /* Setup input buffer */
    assert( p_block->i_buffer <= p_sys->in.bufSizes[0] );

    if( i_extra > 0 )
    {
        memcpy( p_sys->in.bufs[0], p_dec->fmt_in.p_extra, i_extra );
    }
    memcpy( p_sys->in.bufs[0] + i_extra, p_block->p_buffer,
            in_args.numBytes - i_extra );

#ifdef DEBUG_DAVINCI
    msg_Dbg( p_dec, "Frame is %d bytes (%d extra), %dx%d, type %s",
            (int)in_args.numBytes, i_extra, (int)status.outputWidth,
            (int)status.outputHeight,
              status.contentType == IVIDEO_PROGRESSIVE ? "progressive" :
              status.contentType == IVIDEO_INTERLACED ? "interlaced" :
              "unknown" );
#endif

    /* Setup output arguemnts */
    out_args.size = sizeof( out_args );

    /* Decode the video */
    if( ( i = VIDDEC_process( p_sys->d, &p_sys->in, &p_sys->out, &in_args, &out_args ) )
        != VIDDEC_EOK )
    {
        msg_Err( p_dec, "Video decoding failed: %s",
            davinci_GetExtendedError(out_args.extendedError ) );

#if 0
        if( (unsigned int)out_args.extendedError == 0x402 )
        {
            printf("frame was %d bytes\n",(unsigned int)in_args.numBytes);
            unsigned int j;
            for(j=0;j<(unsigned int)in_args.numBytes;j++)
                printf("%.2x ",p_sys->in.bufs[0][j]&0xff);
            printf("\n");

            CERuntime_exit();
            abort();
        }
#endif
        if( i_extra == 0 && p_dec->fmt_in.i_extra > 0 )
        {   /* FIXME : make sure this is correct with the API */
            msg_Dbg( p_dec, "Trying again with p_extra" );
            return DecodeVideoBlockInner( p_dec, pp_block, p_dec->fmt_in.i_extra );
        }
        goto error;
    }

    if( VIDDEC_control( p_sys->d, XDM_GETSTATUS, &dparams, &status ) != VIDDEC_EOK )
    {
        msg_Err( p_dec, "Failed to get decoder status" );
        goto error;
    }

    if( in_args.numBytes > out_args.bytesConsumed )
        msg_Err( p_dec, "%d bytes were not consumed", (int)(out_args.bytesConsumed - in_args.numBytes) );

    p_block->p_buffer += out_args.bytesConsumed - i_extra;
    p_block->i_buffer -= out_args.bytesConsumed - i_extra;

    p_dec->fmt_out.video.i_width = status.outputWidth;
    p_dec->fmt_out.video.i_height = status.outputHeight;
    p_dec->fmt_out.video.i_aspect = VOUT_ASPECT_FACTOR * status.outputWidth
                                    / status.outputHeight; /* FIXME */

    /* Get a new picture */
    p_pic = p_dec->pf_vout_buffer_new( p_dec );
    if( !p_pic )
    {
        msg_Err( p_dec, "Could not get picture" );
        goto error;
    }
    p_pic->b_progressive = status.contentType == IVIDEO_INTERLACED ? VLC_FALSE
                                                                   : VLC_TRUE;

    /* Copy stuff to picture here */
#ifdef DEBUG_DAVINCI
    switch( out_args.decodedFrameType )
    {
        case IVIDEO_NA_FRAME:
            msg_Dbg( p_dec, "Unknown frame type" );
            break;
        case IVIDEO_I_FRAME:
            msg_Dbg( p_dec, "I frame" );
            break;
        case IVIDEO_P_FRAME:
            msg_Dbg( p_dec, "P frame" );
            break;
        case IVIDEO_B_FRAME:
            msg_Dbg( p_dec, "B frame" );
            break;
        case IVIDEO_II_FRAME:
            msg_Dbg( p_dec, "Interlaced I frame" );
            break;
        case IVIDEO_PP_FRAME:
            msg_Dbg( p_dec, "Interlaced P frame" );
            break;
        case IVIDEO_BB_FRAME:
            msg_Dbg( p_dec, "Interlaced B frame" );
            break;
        default:
            msg_Dbg( p_dec, "Other frame type: %d", (int)out_args.decodedFrameType );
            break;
    }
#endif

    assert( p_pic->i_planes == p_sys->out.numBufs );

    for( i = 0; i < p_pic->i_planes; i++ )
    {
        plane_t *p = p_pic->p+i;
        memcpy( p->p_pixels, p_sys->out.bufs[i], p->i_pitch * p->i_visible_lines );
    }

    p_pic->date = p_block->i_pts > 0 ? p_block->i_pts : p_block->i_dts;

    if( p_block->i_buffer == 0 )
    {
        block_Release( p_block );
        *pp_block = NULL;
    }

    return p_pic;

error:
    if( p_pic && p_pic->pf_release )
        p_pic->pf_release( p_pic );
    /* Do we really need this everytime? (like if buffer memory allocation fails we could keep the block) */
    block_Release( p_block );
    return NULL;
}

