/*****************************************************************************
 * videnc.c: video encoder module using the DaVinci DSP.
 *****************************************************************************
 * Copyright (C) 2008 M2X BV
 * $Id: 484dd9911d31c9eef0670b8571fb78d036ae4aae $
 *
 * Authors: Antoine Cellerier <dionoea at videolan dot org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston MA 02110-1301, USA.
 *****************************************************************************/

/*****************************************************************************
 * Preamble
 *****************************************************************************/

#include "davinci.h"
#include <assert.h>
#include <ti/sdo/ce/video/videnc.h>

/*****************************************************************************
 * Local prototypes
 *****************************************************************************/
static block_t *EncodeVideo( encoder_t *, picture_t * );

/*****************************************************************************
 * Video encoder
 *****************************************************************************/
struct encoder_sys_t
{
    Engine_Handle e;
    VIDENC_Handle c;

    XDM_BufDesc in;
    XDM_BufDesc out;
};

/*****************************************************************************
 *
 *****************************************************************************/
int OpenVideoEncoder( vlc_object_t *p_this )
{
    encoder_t *p_enc = (encoder_t *)p_this;
    encoder_sys_t *p_sys;
    char *psz_codec = NULL;
    Engine_Error err;
    VIDENC_Params params;

    switch( p_enc->fmt_out.i_codec )
    {
#if 0
        CASE_MPEG1
            psz_codec = strdup( "mpegenc" ); /* FIXME */
            break;

        CASE_MPEG2
            psz_codec = strdup( "mpeg2enc" ); /* FIXME */
            break;
#endif

        CASE_MPEG4
            psz_codec = strdup( "mpeg4enc" );
            break;

        CASE_H264
            psz_codec = strdup( "h264enc" );
            break;

#if 0
        CASE_VC1
            psz_codec = strdup( "vc1enc" ); /* FIXME */
            break;
#endif

        default:
            return VLC_EGENERIC;
    }

    /* Allocate our private structure */
    p_enc->p_sys = (encoder_sys_t *)malloc( sizeof( encoder_sys_t ) );
    if( !p_enc->p_sys )
        return VLC_ENOMEM;
    p_sys = p_enc->p_sys;
    memset( p_sys, 0, sizeof( encoder_sys_t ) );

    /* Initialize the codec engine */
    CERuntime_init();

    /* Create an engine handle */
    #define ti_engine "encode"
    p_sys->e = Engine_open( ti_engine, NULL /*&Engine_ATTRS*/, &err );
    if( err != Engine_EOK )
    {
        msg_Err( p_enc, "Error while opening engine `%s': %s",
                 ti_engine, ppsz_engine_error[err] );
        goto error;
    }

    PrintAvailableAlgorithms( p_this, ti_engine );

    /* Configure encoder */
    params.size = sizeof( params );
    params.encodingPreset = 0; /* Use default */
    params.rateControlPreset = 0; /* Use default */
    params.maxHeight = p_enc->fmt_in.video.i_height;
    params.maxWidth = p_enc->fmt_in.video.i_width;
    params.maxFrameRate = p_enc->fmt_out.video.i_frame_rate_base ?
                          p_enc->fmt_out.video.i_frame_rate * 1000 /
                          p_enc->fmt_out.video.i_frame_rate_base :
                          25000; /* Frames per 1000 seconds  */
    params.maxBitRate = p_enc->fmt_out.i_bitrate
                      + p_enc->i_tolerance; /* Bits per second. FIXME? */
    params.maxBitRate *= 10;
    params.dataEndianness = XDM_BYTE;
    params.maxInterFrameInterval = p_enc->i_iframes; /* FIXME? */
    params.inputChromaFormat = VlcChromaToXdm( p_enc->fmt_in.i_codec );
    msg_Dbg( p_enc, "%dx%d at %.3f fps (max bitrate %d kBps, I-Frame interval %d)\n",
        (int)params.maxHeight, (int)params.maxWidth, 
        ((float)params.maxFrameRate)/1000.,
        ((int)params.maxBitRate) / (8*1024),
        (int)params.maxInterFrameInterval
            );
    if( params.inputChromaFormat == XDM_CHROMA_NA )
    {
        msg_Warn( p_enc, "Unsupported input chroma (%4.4s), forcing I420.",
                 (const char *)&p_enc->fmt_in.i_codec );
        params.inputChromaFormat = VlcChromaToXdm( VLC_FOURCC('I','4','2','0' ) );
    }
    params.inputContentType = IVIDEO_PROGRESSIVE; /* FIXME: we don't know if it's progressive or interlaced until we get the first picture (p_pic->b_progressive) */
    /* Create encoder handle */
    p_sys->c = VIDENC_create( p_sys->e, (String)psz_codec, &params );
    if( !p_sys->c )
    {
        msg_Err( p_enc, "Failed to create video encoder (%s)", psz_codec );
        goto error;
    }

    /* Initialize random stuff */
    p_enc->pf_encode_video = EncodeVideo;

    free( psz_codec );
    return VLC_SUCCESS;

    error:
        if( p_sys->e ) Engine_close( p_sys->e );
        free( p_sys );
        free( psz_codec );
        return VLC_EGENERIC;
}

/*****************************************************************************
 *
 *****************************************************************************/
void CloseVideoEncoder( vlc_object_t *p_this )
{
    encoder_t *p_enc = (encoder_t *)p_this;
    encoder_sys_t *p_sys = p_enc->p_sys;

    /* Delete video encoder */
    VIDENC_delete( p_sys->c );

    /* Delete codec engine handle */
    Engine_close( p_sys->e );

    /* Exit the codec engine */
    CERuntime_exit(); /* XXX: Should this be done only once (if we have encoder + decoder runing) */

    /* Free 'DaVinci compliant' buffers */
    FreeBuffer( &p_sys->in );
    FreeBuffer( &p_sys->out );

    free( p_sys );
}

/*****************************************************************************
 *
 *****************************************************************************/
static block_t *EncodeVideo( encoder_t *p_enc, picture_t *p_pic )
{
    encoder_sys_t *p_sys = p_enc->p_sys;
    block_t *p_block;
    int i;
    VIDENC_InArgs in_args;
    VIDENC_OutArgs out_args;

    if( p_sys->in.numBufs == 0 || p_sys->out.numBufs == 0 )
    {
        VIDENC_DynamicParams dparams;
        VIDENC_Status status;

        dparams.size = sizeof( dparams );
        memset( &status, 0, sizeof( status ) );
        status.size = sizeof( status );

        /* Configue the encoder */
        dparams.inputHeight = p_enc->fmt_in.video.i_height;
        dparams.inputWidth = p_enc->fmt_in.video.i_width;
        dparams.refFrameRate = p_enc->fmt_out.video.i_frame_rate_base ?
                               p_enc->fmt_out.video.i_frame_rate * 1000 /
                               p_enc->fmt_out.video.i_frame_rate_base :
                               30; /* Frames per 1000 seconds  */
        dparams.targetFrameRate = dparams.refFrameRate; /* FIXME? */
        dparams.targetBitRate = p_enc->fmt_out.i_bitrate;
        dparams.intraFrameInterval = p_enc->i_iframes;
        dparams.generateHeader = XDM_ENCODE_AU; /* FIXME? */
        dparams.captureWidth = p_pic->p->i_pitch / p_pic->p->i_pixel_pitch; /* FIXME? */
        dparams.forceIFrame = 0;
        if( VIDENC_control( p_sys->c, XDM_SETPARAMS, &dparams, &status )
            != VIDENC_EOK )
        {
            msg_Err( p_enc, "Failed to set encoder parameters" );
            goto error;
        }

        /* Configure buffers */
        if( VIDENC_control( p_sys->c, XDM_GETBUFINFO, &dparams, &status )
            != VIDENC_EOK )
        {
            msg_Err( p_enc, "Failed to get buffer info" );
            goto error;
        }

        /* Allocate input buffer(s) */
        if( AllocateBuffer( p_enc, status.bufInfo.minNumInBufs,
                            status.bufInfo.minInBufSize, &p_sys->in )
            != VLC_SUCCESS )
        {
            msg_Err( p_enc, "Failed to allocate input buffers" );
            goto error;
        }

        /* Allocate output buffer */
        if( AllocateBuffer( p_enc, status.bufInfo.minNumOutBufs,
                            status.bufInfo.minOutBufSize, &p_sys->out )
            != VLC_SUCCESS )
        {
            msg_Err( p_enc, "Failed to allocate input buffers" );
            goto error;
        }
    }

    /* Copy input picture */
    assert( p_pic->i_planes == p_sys->in.numBufs );

    for( i = 0; i < p_pic->i_planes; i++ )
    {
        plane_t *p = p_pic->p+i;
        memcpy( p_sys->in.bufs[i], p->p_pixels, p->i_pitch * p->i_visible_lines );
    }

    /* Configure input */
    in_args.size = sizeof( in_args );

    /* Configure output */
    memset( &out_args, 0, sizeof( out_args ) );
    out_args.size = sizeof( out_args );

    /* Encode the video */
    i = VIDENC_process( p_sys->c, &p_sys->in, &p_sys->out,
                        &in_args, &out_args );
    if( i != VIDENC_EOK )
    {
        msg_Err( p_enc, "Video encoding failed: %s",
                 davinci_GetExtendedError( out_args.extendedError ) );
        goto error;
    }

    /* Print some info */
    msg_Dbg( p_enc, "Bytes generated: %d", (int)out_args.bytesGenerated );

    /* Put everything in the block */
    if( out_args.bytesGenerated <= 0 )
        goto error;

    p_block = block_New( p_enc, out_args.bytesGenerated );
    if( !p_block ) goto error;

    memcpy( p_block->p_buffer, p_sys->out.bufs[0], out_args.bytesGenerated );

    switch( out_args.encodedFrameType )
    {
        case IVIDEO_I_FRAME:
        case IVIDEO_II_FRAME:
            p_block->i_flags |= BLOCK_FLAG_TYPE_I;
            break;
        case IVIDEO_P_FRAME:
        case IVIDEO_PP_FRAME:
            p_block->i_flags |= BLOCK_FLAG_TYPE_P;
            break;
        case IVIDEO_B_FRAME:
        case IVIDEO_BB_FRAME:
            p_block->i_flags |= BLOCK_FLAG_TYPE_B;
            break;
        default:
            msg_Warn( p_enc, "Unknown frame type" );
            p_block->i_flags |= BLOCK_FLAG_TYPE_PB; /* Wild guess */
            break;
    }

    p_block->i_dts = p_block->i_pts = p_pic->date;

    /* Shamelessly copied from ffmpeg/encoder.c */
    p_block->i_length = I64C(1000000) *
        p_enc->fmt_in.video.i_frame_rate_base /
            p_enc->fmt_in.video.i_frame_rate;

    return p_block;

    error:
        return NULL;
}
